/*
 * Sixth 3D engine. Author: Svjatoslav Agejenko. 
 * This project is released under Creative Commons Zero (CC0) license.
 */
package eu.svjatoslav.sixth.e3d.renderer.octree.raytracer;

import eu.svjatoslav.sixth.e3d.gui.ViewPanel;
import eu.svjatoslav.sixth.e3d.renderer.octree.OctreeVolume;
import eu.svjatoslav.sixth.e3d.renderer.raster.Color;
import eu.svjatoslav.sixth.e3d.renderer.raster.texture.Texture;

import java.util.Vector;

public class RayTracer implements Runnable {

    private static final int PROGRESS_UPDATE_FREQUENCY_MILLIS = 1000;
    private final Camera camera;
    private final Texture texture;
    private final ViewPanel viewPanel;
    private OctreeVolume octreeVolume;
    private Vector<LightSource> lights;
    private int computedLights;

    public RayTracer(final Texture texture, final OctreeVolume octreeVolume,
                     final Vector<LightSource> lights, final Camera camera,
                     final ViewPanel viewPanel) {

        this.texture = texture;
        this.octreeVolume = octreeVolume;
        this.lights = lights;
        this.camera = camera;
        this.viewPanel = viewPanel;
    }

    @Override
    public void run() {
        computedLights = 0;

        // create camera

        // Camera cam = new Camera(camCenter, upLeft, upRight, downLeft,
        // downRight);

        // add camera to the raytracing point
        // Main.mainWorld.geometryCollection.addObject(cam);
        // Main.mainWorld.compiledGeometry.compileGeometry(Main.mainWorld.geometryCollection);

        final int width = texture.primaryBitmap.width;
        final int height = texture.primaryBitmap.height;

        final CameraView cameraView = camera.getCameraView();

        // calculate vertical vectors
        final double x1p = cameraView.downLeft.x - cameraView.upLeft.x;
        final double y1p = cameraView.downLeft.y - cameraView.upLeft.y;
        final double z1p = cameraView.downLeft.z - cameraView.upLeft.z;

        final double x2p = cameraView.downRight.x - cameraView.upRight.x;
        final double y2p = cameraView.downRight.y - cameraView.upRight.y;
        final double z2p = cameraView.downRight.z - cameraView.upRight.z;

        long nextBitmapUpdate = System.currentTimeMillis()
                + PROGRESS_UPDATE_FREQUENCY_MILLIS;

        for (int y = 0; y < height; y++) {
            final double cx1 = cameraView.upLeft.x + ((x1p * y) / height);
            final double cy1 = cameraView.upLeft.y + ((y1p * y) / height);
            final double cz1 = cameraView.upLeft.z + ((z1p * y) / height);

            final double cx2 = cameraView.upRight.x + ((x2p * y) / height);
            final double cy2 = cameraView.upRight.y + ((y2p * y) / height);
            final double cz2 = cameraView.upRight.z + ((z2p * y) / height);

            // calculate horisontal vector
            final double x3p = cx2 - cx1;
            final double y3p = cy2 - cy1;
            final double z3p = cz2 - cz1;

            for (int x = 0; x < width; x++) {
                final double cx3 = cx1 + ((x3p * x) / width);
                final double cy3 = cy1 + ((y3p * x) / width);
                final double cz3 = cz1 + ((z3p * x) / width);

                final Ray r = new Ray(cameraView.camCenter.x,
                        cameraView.camCenter.y, cameraView.camCenter.z, cx3
                        - cameraView.camCenter.x, cy3
                        - cameraView.camCenter.y, cz3
                        - cameraView.camCenter.z);
                final int c = traceRay(r);

                final Color color = new Color(c);
                texture.primaryBitmap.drawPixel(x, y, color);
            }

            if (System.currentTimeMillis() > nextBitmapUpdate) {
                nextBitmapUpdate = System.currentTimeMillis()
                        + PROGRESS_UPDATE_FREQUENCY_MILLIS;
                texture.resetResampledBitmapCache();
                viewPanel.repaintDuringNextViewUpdate();
            }
        }

        texture.resetResampledBitmapCache();
        viewPanel.repaintDuringNextViewUpdate();
    }

    private int traceRay(final Ray ray) {

        final int intersectingCell = octreeVolume.traceCell(0, 0, 0,
                octreeVolume.masterCellSize, 0, ray);

        if (intersectingCell != -1) {
            // if lightening not computed, compute it
            if (octreeVolume.ce3[intersectingCell] == -1)
                // if cell is larger than 1
                if (ray.hitCellSize > 1) {
                    // break it up
                    octreeVolume.breakSolidCell(intersectingCell);
                    return traceRay(ray);
                } else {
                    computedLights++;
                    float red = 30, green = 30, blue = 30;

                    for (final LightSource l : lights) {
                        final int xDist = (l.x - ray.hitCellX);
                        final int yDist = (l.y - ray.hitCellY);
                        final int zDist = (l.z - ray.hitCellZ);

                        double newRed = 0, newGreen = 0, newBlue = 0;
                        double tempRed, tempGreen, tempBlue;

                        double distance = Math.sqrt((xDist * xDist)
                                + (yDist * yDist) + (zDist * zDist));
                        distance = (distance / 3) + 1;

                        final Ray r1 = new Ray(ray.hitCellX, ray.hitCellY
                                - (float) 1.5, ray.hitCellZ,

                                (float) l.x - (float) ray.hitCellX, l.y
                                - (ray.hitCellY - (float) 1.5), (float) l.z
                                - (float) ray.hitCellZ);

                        final int rt1 = octreeVolume.traceCell(0, 0, 0,
                                octreeVolume.masterCellSize, 0, r1);

                        if (rt1 == -1) {
                            newRed = (l.color.r * l.brightness) / distance;
                            newGreen = (l.color.g * l.brightness) / distance;
                            newBlue = (l.color.b * l.brightness) / distance;
                        }

                        final Ray r2 = new Ray(ray.hitCellX - (float) 1.5,
                                ray.hitCellY, ray.hitCellZ,

                                l.x - (ray.hitCellX - (float) 1.5), (float) l.y
                                - (float) ray.hitCellY, (float) l.z
                                - (float) ray.hitCellZ);

                        final int rt2 = octreeVolume.traceCell(0, 0, 0,
                                octreeVolume.masterCellSize, 0, r2);

                        if (rt2 == -1) {
                            tempRed = (l.color.r * l.brightness) / distance;
                            tempGreen = (l.color.g * l.brightness) / distance;
                            tempBlue = (l.color.b * l.brightness) / distance;

                            if (tempRed > newRed)
                                newRed = tempRed;
                            if (tempGreen > newGreen)
                                newGreen = tempGreen;
                            if (tempBlue > newBlue)
                                newBlue = tempBlue;
                        }

                        final Ray r3 = new Ray(ray.hitCellX, ray.hitCellY,
                                ray.hitCellZ - (float) 1.5,

                                (float) l.x - (float) ray.hitCellX, (float) l.y
                                - (float) ray.hitCellY, l.z
                                - (ray.hitCellZ - (float) 1.5));

                        final int rt3 = octreeVolume.traceCell(0, 0, 0,
                                octreeVolume.masterCellSize, 0, r3);

                        if (rt3 == -1) {
                            tempRed = (l.color.r * l.brightness) / distance;
                            tempGreen = (l.color.g * l.brightness) / distance;
                            tempBlue = (l.color.b * l.brightness) / distance;
                            if (tempRed > newRed)
                                newRed = tempRed;
                            if (tempGreen > newGreen)
                                newGreen = tempGreen;
                            if (tempBlue > newBlue)
                                newBlue = tempBlue;
                        }

                        final Ray r4 = new Ray(ray.hitCellX, ray.hitCellY
                                + (float) 1.5, ray.hitCellZ,

                                (float) l.x - (float) ray.hitCellX, l.y
                                - (ray.hitCellY + (float) 1.5), (float) l.z
                                - (float) ray.hitCellZ);

                        final int rt4 = octreeVolume.traceCell(0, 0, 0,
                                octreeVolume.masterCellSize, 0, r4);

                        if (rt4 == -1) {
                            tempRed = (l.color.r * l.brightness) / distance;
                            tempGreen = (l.color.g * l.brightness) / distance;
                            tempBlue = (l.color.b * l.brightness) / distance;
                            if (tempRed > newRed)
                                newRed = tempRed;
                            if (tempGreen > newGreen)
                                newGreen = tempGreen;
                            if (tempBlue > newBlue)
                                newBlue = tempBlue;
                        }

                        final Ray r5 = new Ray(ray.hitCellX + (float) 1.5,
                                ray.hitCellY, ray.hitCellZ,

                                l.x - (ray.hitCellX + (float) 1.5), (float) l.y
                                - (float) ray.hitCellY, (float) l.z
                                - (float) ray.hitCellZ);

                        final int rt5 = octreeVolume.traceCell(0, 0, 0,
                                octreeVolume.masterCellSize, 0, r5);

                        if (rt5 == -1) {
                            tempRed = (l.color.r * l.brightness) / distance;
                            tempGreen = (l.color.g * l.brightness) / distance;
                            tempBlue = (l.color.b * l.brightness) / distance;
                            if (tempRed > newRed)
                                newRed = tempRed;
                            if (tempGreen > newGreen)
                                newGreen = tempGreen;
                            if (tempBlue > newBlue)
                                newBlue = tempBlue;
                        }

                        final Ray r6 = new Ray(ray.hitCellX, ray.hitCellY,
                                ray.hitCellZ + (float) 1.5,

                                (float) l.x - (float) ray.hitCellX, (float) l.y
                                - (float) ray.hitCellY, l.z
                                - (ray.hitCellZ + (float) 1.5));

                        final int rt6 = octreeVolume.traceCell(0, 0, 0,
                                octreeVolume.masterCellSize, 0, r6);

                        if (rt6 == -1) {
                            tempRed = (l.color.r * l.brightness) / distance;
                            tempGreen = (l.color.g * l.brightness) / distance;
                            tempBlue = (l.color.b * l.brightness) / distance;
                            if (tempRed > newRed)
                                newRed = tempRed;
                            if (tempGreen > newGreen)
                                newGreen = tempGreen;
                            if (tempBlue > newBlue)
                                newBlue = tempBlue;
                        }
                        red += newRed;
                        green += newGreen;
                        blue += newBlue;

                    }

                    final int cellColor = octreeVolume.ce2[intersectingCell];

                    red = (red * ((cellColor & 0xFF0000) >> 16)) / 255;
                    green = (green * ((cellColor & 0xFF00) >> 8)) / 255;
                    blue = (blue * (cellColor & 0xFF)) / 255;

                    if (red > 255)
                        red = 255;
                    if (green > 255)
                        green = 255;
                    if (blue > 255)
                        blue = 255;

                    octreeVolume.ce3[intersectingCell] = (((int) red) << 16)
                            + (((int) green) << 8) + ((int) blue);

                }
            if (octreeVolume.ce3[intersectingCell] == 0)
                return octreeVolume.ce2[intersectingCell];
            return octreeVolume.ce3[intersectingCell];
        }

        // return (200 << 16) + (200 << 8) + 255;
        return 0;
    }

}
