/*
 * Sixth 3D engine. Author: Svjatoslav Agejenko. 
 * This project is released under Creative Commons Zero (CC0) license.
 */
package eu.svjatoslav.sixth.e3d.renderer.raster.shapes.composite.textcanvas;

import eu.svjatoslav.sixth.e3d.geometry.Point3D;
import eu.svjatoslav.sixth.e3d.gui.RenderingContext;
import eu.svjatoslav.sixth.e3d.gui.TextPointer;
import eu.svjatoslav.sixth.e3d.math.Transform;
import eu.svjatoslav.sixth.e3d.math.TransformsStack;
import eu.svjatoslav.sixth.e3d.renderer.raster.Color;
import eu.svjatoslav.sixth.e3d.renderer.raster.shapes.composite.TexturedRectangle;

import java.awt.*;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;

import static eu.svjatoslav.sixth.e3d.renderer.raster.Color.BLACK;
import static eu.svjatoslav.sixth.e3d.renderer.raster.Color.WHITE;

public class TextCanvas extends TexturedRectangle {

    // character size in world coordinates
    public static final int FONT_CHAR_WIDTH = 8;
    public static final int FONT_CHAR_HEIGHT = 16;

    // character size on texture
    public static final int FONT_CHAR_WIDTH_TEXTURE_PIXELS = 16;
    public static final int FONT_CHAR_HEIGHT_TEXTURE_PIXELS = 32;


    public static final Font FONT = CanvasCharacter.getFont((int)(FONT_CHAR_HEIGHT_TEXTURE_PIXELS / 1.066));
    private static final String GROUP_TEXTURE = "texture";
    private static final String GROUP_CHARACTERS = "characters";
    private final TextPointer size;
    private final TextPointer cursorLocation = new TextPointer();
    CanvasCharacter[][] lines;
    private RenderMode renderMode = null;
    private Color backgroundColor = BLACK;
    private Color foregroundColor = WHITE;

    public TextCanvas(final Transform location, final String text,
                      final Color foregroundColor, final Color backgroundColor) {
        this(location, getTextDimensions(text), foregroundColor,
                backgroundColor);
        setText(text);
    }

    public TextCanvas(final Transform location, final TextPointer dimensions,
                      final Color foregroundColor, final Color backgroundColor) {
        super(location);
        getRelativityTracker().enableOrientationTracking();

        size = dimensions;
        final int columns = dimensions.column;
        final int rows = dimensions.row;

        this.backgroundColor = backgroundColor;
        this.foregroundColor = foregroundColor;

        // initialize underlying textured rectangle
        initialize(
                columns * FONT_CHAR_WIDTH,
                rows * FONT_CHAR_HEIGHT,
                columns * FONT_CHAR_WIDTH_TEXTURE_PIXELS,
                rows * FONT_CHAR_HEIGHT_TEXTURE_PIXELS,
                0);

        getTexture().primaryBitmap.fillColor(backgroundColor);
        getTexture().resetResampledBitmapCache();

        setGroupForUngrouped(GROUP_TEXTURE);

        lines = new CanvasCharacter[rows][];
        for (int row = 0; row < rows; row++) {
            lines[row] = new CanvasCharacter[columns];

            for (int column = 0; column < columns; column++) {
                final Point3D characterCoordinate = getCharLocation(row, column);

                final CanvasCharacter character = new CanvasCharacter(
                        characterCoordinate, ' ', foregroundColor,
                        backgroundColor);
                addShape(character);
                lines[row][column] = character;
            }

        }

        setGroupForUngrouped(GROUP_CHARACTERS);

        setRenderMode(RenderMode.TEXTURE);
    }

    public static TextPointer getTextDimensions(final String text) {

        final BufferedReader reader = new BufferedReader(new StringReader(text));

        int rows = 0;
        int columns = 0;

        while (true) {
            final String line;
            try {
                line = reader.readLine();
            } catch (IOException e) {
                throw new RuntimeException(e);
            }

            if (line == null)
                return new TextPointer(rows, columns);

            rows++;
            columns = Math.max(columns, line.length());
        }
    }

    @Override
    public void beforeTransformHook(final TransformsStack transformPipe,
                                    final RenderingContext context) {

        final double textRelativeSize = context.width
                / getRelativityTracker().getDistanceToUser();

        if (textRelativeSize < 2d) {
            if (renderMode == RenderMode.CHARACTERS)
                setRenderMode(RenderMode.TEXTURE);
            return;
        }

        final double piHalf = Math.PI / 2;

        final double deviation = Math.abs(getRelativityTracker().getAngleXZ()
                + piHalf)
                + Math.abs(getRelativityTracker().getAngleYZ() + piHalf);

        final double maxDeviation = 0.5;

        if (deviation > maxDeviation) {
            if (renderMode == RenderMode.CHARACTERS)
                setRenderMode(RenderMode.TEXTURE);
        } else if (renderMode == RenderMode.TEXTURE)
            setRenderMode(RenderMode.CHARACTERS);
    }

    public void clear() {
        for (final CanvasCharacter[] line : lines)
            for (final CanvasCharacter character : line) {
                character.setValue(' ');
                character.setBackgroundColor(backgroundColor);
                character.setForegroundColor(foregroundColor);
            }

        // set background color
        getTexture().primaryBitmap.fillColor(backgroundColor);
        getTexture().resetResampledBitmapCache();
    }

    private void drawCharToTexture(final int row, final int column,
                                   final char character, final Color foreground) {
        final Graphics2D graphics = getTexture().graphics;

        getTexture().primaryBitmap.drawRectangle(
                column * FONT_CHAR_WIDTH_TEXTURE_PIXELS,
                row * FONT_CHAR_HEIGHT_TEXTURE_PIXELS,
                (column * FONT_CHAR_WIDTH_TEXTURE_PIXELS) + FONT_CHAR_WIDTH_TEXTURE_PIXELS,
                (row * FONT_CHAR_HEIGHT_TEXTURE_PIXELS) + FONT_CHAR_HEIGHT_TEXTURE_PIXELS,
                backgroundColor);

        graphics.setFont(FONT);
        graphics.setColor(foreground.toAwtColor());
        graphics.drawChars(
                new char[]{character,}, 0, 1,
                (column * FONT_CHAR_WIDTH_TEXTURE_PIXELS) - 0,
                (row * FONT_CHAR_HEIGHT_TEXTURE_PIXELS) + (int)(FONT_CHAR_HEIGHT_TEXTURE_PIXELS / 1.23f));

        getTexture().resetResampledBitmapCache();
    }

    public Point3D getCharLocation(final int row, final int column) {
        final Point3D coordinate = topLeft.clone();

        coordinate.translateY((row * FONT_CHAR_HEIGHT)
                + (FONT_CHAR_HEIGHT / 3.2));

        coordinate.translateX((column * FONT_CHAR_WIDTH)
                + (FONT_CHAR_WIDTH / 2));

        return coordinate;
    }

    public TextPointer getSize() {
        return size;
    }

    public void locate(final int row, final int column) {
        cursorLocation.row = row;
        cursorLocation.column = column;
    }

    public void print(final String text) {
        for (final char c : text.toCharArray())
            putChar(c);
    }

    public void putChar(final char character) {
        putChar(cursorLocation, character);

        cursorLocation.column++;
        if (cursorLocation.column >= size.column) {
            cursorLocation.column = 0;
            cursorLocation.row++;
        }
    }

    public void putChar(final int row, final int column, final char character) {
        if ((row >= lines.length) || (row < 0))
            return;

        final CanvasCharacter[] line = lines[row];

        if ((column >= line.length) || (column < 0))
            return;

        final CanvasCharacter canvasCharacter = line[column];
        canvasCharacter.setValue(character);
        canvasCharacter.setBackgroundColor(backgroundColor);
        canvasCharacter.setForegroundColor(foregroundColor);
        drawCharToTexture(row, column, character,
                foregroundColor);
    }

    public void putChar(final TextPointer location, final char character) {
        putChar(location.row, location.column, character);
    }

    public void setBackgroundColor(
            final eu.svjatoslav.sixth.e3d.renderer.raster.Color backgroundColor) {
        this.backgroundColor = backgroundColor;
    }

    public void setForegroundColor(
            final eu.svjatoslav.sixth.e3d.renderer.raster.Color foregroundColor) {
        this.foregroundColor = foregroundColor;
    }

    private void setRenderMode(final RenderMode mode) {
        if (mode == renderMode)
            return;

        switch (mode) {
            case CHARACTERS:
                hideGroup(GROUP_TEXTURE);
                showGroup(GROUP_CHARACTERS);
                break;
            case TEXTURE:
                hideGroup(GROUP_CHARACTERS);
                showGroup(GROUP_TEXTURE);
                break;
        }

        renderMode = mode;
    }

    public void setText(final String text) {
        final BufferedReader reader = new BufferedReader(new StringReader(text));

        int row = 0;

        while (true) {
            final String line;
            try {
                line = reader.readLine();
            } catch (IOException e) {
                throw new RuntimeException(e);
            }

            if (line == null)
                return;

            int column = 0;
            for (final char c : line.toCharArray()) {
                putChar(row, column, c);
                column++;
            }
            row++;
        }
    }

    public void setTextColor(final Color color) {
        for (final CanvasCharacter[] line : lines)
            for (final CanvasCharacter character : line)
                character.setForegroundColor(color);
    }

}
