/*
 * Sixth 3D engine. Author: Svjatoslav Agejenko. 
 * This project is released under Creative Commons Zero (CC0) license.
 */
package eu.svjatoslav.sixth.e3d.renderer.raster.texture;

import eu.svjatoslav.sixth.e3d.gui.RenderingContext;

import java.awt.*;
import java.awt.image.BufferedImage;
import java.awt.image.DataBufferByte;
import java.awt.image.WritableRaster;

import static java.util.Arrays.fill;

public class Texture {

    public final TextureBitmap primaryBitmap;
    public final java.awt.Graphics2D graphics;
    TextureBitmap[] upSampled;
    TextureBitmap[] downSampled = new TextureBitmap[8];

    public Texture(final int width, final int height, final int maxUpscale) {
        upSampled = new TextureBitmap[maxUpscale];

        final BufferedImage bufferedImage = new BufferedImage(width, height,
                RenderingContext.bufferedImageType);

        final WritableRaster raster = bufferedImage.getRaster();
        final DataBufferByte dbi = (DataBufferByte) raster.getDataBuffer();
        graphics = (Graphics2D) bufferedImage.getGraphics();

        graphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                RenderingHints.VALUE_ANTIALIAS_ON);

        graphics.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,
                RenderingHints.VALUE_TEXT_ANTIALIAS_ON);

        primaryBitmap = new TextureBitmap(width, height, dbi.getData(), 1);
    }

    public int detectDownscaleFactorForZoom(final double zoom) {
        double size = 1;
        for (int i = 0; i < downSampled.length; i++) {
            size = size / 2;
            if (size < zoom)
                return i;
        }

        return downSampled.length - 1;
    }

    public int detectUpscaleFactorForZoom(final double zoom) {
        double size = 2;
        for (int i = 0; i < upSampled.length; i++) {
            size = size * 2;
            if (size > zoom)
                return i;
        }

        return upSampled.length - 1;
    }

    /**
     * Downscale given bitmap by factor of 2.
     *
     * @param originalBitmap Bitmap to downscale.
     * @return Downscaled bitmap.
     */
    public TextureBitmap downscaleBitmap(final TextureBitmap originalBitmap) {
        int newWidth = originalBitmap.width / 2;
        int newHeight = originalBitmap.height / 2;

        // Enforce minimum width and height
        if (newWidth < 1)
            newWidth = 1;
        if (newHeight < 1)
            newHeight = 1;

        final TextureBitmap downScaled = new TextureBitmap(newWidth, newHeight,
                originalBitmap.multiplicationFactor / 2d);

        final ColorAccumulator accumulator = new ColorAccumulator();

        for (int y = 0; y < newHeight; y++)
            for (int x = 0; x < newWidth; x++) {
                accumulator.reset();
                accumulator.accumulate(originalBitmap, x * 2, y * 2);
                accumulator.accumulate(originalBitmap, (x * 2) + 1, y * 2);
                accumulator.accumulate(originalBitmap, x * 2, (y * 2) + 1);
                accumulator
                        .accumulate(originalBitmap, (x * 2) + 1, (y * 2) + 1);
                accumulator.storeResult(downScaled, x, y);
            }

        return downScaled;
    }

    public TextureBitmap getDownscaledBitmap(final int scaleFactor) {
        if (downSampled[scaleFactor] == null) {

            TextureBitmap largerBitmap;
            if (scaleFactor == 0)
                largerBitmap = primaryBitmap;
            else
                largerBitmap = getDownscaledBitmap(scaleFactor - 1);

            downSampled[scaleFactor] = downscaleBitmap(largerBitmap);
        }

        return downSampled[scaleFactor];
    }

    /**
     * Returns the bitmap that should be used for rendering at the given zoom
     * @param scaleFactor The upscale factor
     * @return The bitmap
     */
    public TextureBitmap getUpscaledBitmap(final int scaleFactor) {
        if (upSampled[scaleFactor] == null) {

            TextureBitmap smallerBitmap;
            if (scaleFactor == 0)
                smallerBitmap = primaryBitmap;
            else
                smallerBitmap = getUpscaledBitmap(scaleFactor - 1);

            upSampled[scaleFactor] = upscaleBitmap(smallerBitmap);
        }

        return upSampled[scaleFactor];
    }

    /**
     * Returns the bitmap that should be used for rendering at the given zoom
     * @param zoomLevel The zoom level
     * @return The bitmap
     */
    public TextureBitmap getZoomedBitmap(final double zoomLevel) {

        if (zoomLevel < 1) {
            final int downscaleFactor = detectDownscaleFactorForZoom(zoomLevel);
            return getDownscaledBitmap(downscaleFactor);
        } else if (zoomLevel > 2) {
            final int upscaleFactor = detectUpscaleFactorForZoom(zoomLevel);

            if (upscaleFactor < 0)
                return primaryBitmap;

            return getUpscaledBitmap(upscaleFactor);
        }

        // System.out.println(zoomLevel);
        return primaryBitmap;
    }

    /**
     * Resets the cache of resampled bitmaps
     */
    public void resetResampledBitmapCache() {
        fill(upSampled, null);

        fill(downSampled, null);
    }

    /**
     * Upscales the given bitmap by a factor of 2
     * @param originalBitmap The bitmap to upscale
     * @return The upscaled bitmap
     */
    public TextureBitmap upscaleBitmap(final TextureBitmap originalBitmap) {
        final int newWidth = originalBitmap.width * 2;
        final int newHeight = originalBitmap.height * 2;

        final TextureBitmap upScaled = new TextureBitmap(newWidth, newHeight,
                originalBitmap.multiplicationFactor * 2d);

        final ColorAccumulator accumulator = new ColorAccumulator();

        for (int y = 0; y < originalBitmap.height; y++)
            for (int x = 0; x < originalBitmap.width; x++) {
                accumulator.reset();
                accumulator.accumulate(originalBitmap, x, y);
                accumulator.storeResult(upScaled, x * 2, y * 2);

                accumulator.reset();
                accumulator.accumulate(originalBitmap, x, y);
                accumulator.accumulate(originalBitmap, x + 1, y);
                accumulator.storeResult(upScaled, (x * 2) + 1, y * 2);

                accumulator.reset();
                accumulator.accumulate(originalBitmap, x, y);
                accumulator.accumulate(originalBitmap, x, y + 1);
                accumulator.storeResult(upScaled, x * 2, (y * 2) + 1);

                accumulator.reset();
                accumulator.accumulate(originalBitmap, x, y);
                accumulator.accumulate(originalBitmap, x + 1, y);
                accumulator.accumulate(originalBitmap, x, y + 1);
                accumulator.accumulate(originalBitmap, x + 1, y + 1);
                accumulator.storeResult(upScaled, (x * 2) + 1, (y * 2) + 1);
            }

        return upScaled;
    }

    /**
     * A helper class that accumulates color values for a given area of a bitmap
     */
    public static class ColorAccumulator {
        // Accumulated color values
        public int r, g, b, a;

        // Number of pixels that have been accumulated
        public int pixelCount = 0;

        /**
         * Accumulates the color values of the given pixel
         * @param bitmap The bitmap
         * @param x The x coordinate of the pixel
         * @param y The y coordinate of the pixel
         */
        public void accumulate(final TextureBitmap bitmap, final int x,
                               final int y) {
            int address = bitmap.getAddress(x, y);

            a += bitmap.bytes[address] & 0xff;
            address++;

            b += bitmap.bytes[address] & 0xff;
            address++;

            g += bitmap.bytes[address] & 0xff;
            address++;

            r += bitmap.bytes[address] & 0xff;

            pixelCount++;
        }

        /**
         * Resets the accumulator
         */
        public void reset() {
            a = 0;
            r = 0;
            g = 0;
            b = 0;
            pixelCount = 0;
        }

        /**
         * Stores the accumulated color values in the given bitmap
         * @param bitmap The bitmap
         * @param x The x coordinate of the pixel
         * @param y The y coordinate of the pixel
         */
        public void storeResult(final TextureBitmap bitmap, final int x,
                                final int y) {
            int address = bitmap.getAddress(x, y);

            bitmap.bytes[address] = (byte) (a / pixelCount);
            address++;

            bitmap.bytes[address] = (byte) (b / pixelCount);
            address++;

            bitmap.bytes[address] = (byte) (g / pixelCount);
            address++;

            bitmap.bytes[address] = (byte) (r / pixelCount);
        }
    }

}
